//============================================================================
// Name        : GA_02_GraphColoring.cpp
// Author      : Angel Arturo Ramirez Suarez
// Version     : 0.0.1
// Copyright   : By the power of Greyskull!
// Description : Graph Coloring problem genetic algorithm in CPP.
//============================================================================

#include <iostream>
//Used for the rand function.
#include <stdlib.h>
//Used as seed.
#include <time.h>
//For power function.
#include <math.h>
#include <cmath>
//For file reading.
#include <sstream>
#include <string>
#include <fstream>
//To use vectors.
#include<vector>
#include <algorithm>
#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/classification.hpp>

using namespace std;
using namespace boost::algorithm;


//Parameters.
//Population size.
int pobSize = 5;
//Chromosome length.
int lCrom = 5;
//Number of iterations.
int iterSize = 0;
//Best candidate position.
int bestPos = 0;
//Adapation.
float sumAdapt = 0.0;
//Reproduction rate.
float crossRate = 0.0;
//Mutation rate.
float mutRate = 0.0;
//Number of colors.
int nColors = 0;
//Number of vertices.
int nVertex = 0;
//Tolerance.
float tol = 0.0;

const double EULER = 2.7182818284590452;
const double MPI = 3.14159265358979323;

struct TIndividual {
	int *genes;
	float fenotype;
	float adaptation;
	float rel_val;
	float acc_val;
	int conflicts;
};

//To store the best candidates.
TIndividual *bestIndividual = (TIndividual *)malloc(sizeof(TIndividual));
int bestPosition = 0;
int colorNumber = 0;
int iterVal = 0;

//Functions.
void genPop(struct TIndividual *population);
float calcAdapt(struct TIndividual *population, int lCrom);
//vector<string> &getParameters();
void printPop(struct TIndividual *population[]);
int f(struct TIndividual *population);
float decode(int genes[], int lCrom, int nColors);
int bin_ent(int genes[], int lCrom);
bool evalPop(struct TIndividual *population[], int pobSize, float sumAdapt);
void selectPop(struct TIndividual *population[], int pobSize);
void reproduction(struct TIndividual *population[], int pobSize, int lCrom, float crossRate);
void crossPop(struct TIndividual *population, struct TIndividual *population2, struct TIndividual *hijo1, struct TIndividual *hijo2, int lCrom, int punto_cross);
void mutPop(struct TIndividual *population[], int popSize, int lCrom, float mutRate);
void printChromosome(struct TIndividual *population);
int calcChromLength(int nColors, float tol);
vector <std::string> readFile(const char *filename);
bool IsNum(char c);

int main() {
	srand(time(NULL));
	//Get algorithm paramenters.
	cout<<"****************************************************"<<endl;
	cout<<"Loading settings"<<endl;
	cout<<"****************************************************"<<endl;
	//Read the settings file so we don't have to give them as input
	//everytime.
	vector <std::string> settings = readFile("settings.txt");
	for(int i=0;i<settings.size();i++){
		std::string line = settings[i];
		line.erase(std::remove_if(line.begin(), line.end(), &IsNum), line.end());
		if(i == 0)
			pobSize = atoi(line.c_str());
		else if(i == 1)
			iterSize = atoi(line.c_str());
		else if(i == 2)
			crossRate = atof(line.c_str());
		else
			mutRate = atof(line.c_str());
	}
	//Read the problem file to find vertex and constraints.
	vector <std::string> problem = readFile("test.col");
	nVertex = 0;
	nColors = 0;
	//This variable holds the location of the connected solution.
	vector <string> Vertex[problem.size()];
	for(uint i=0;i<problem.size()-3;i++){
		std::string line = problem[i+3];
		//If e found, it means its a vertex.
		if(line.find("e") != string::npos){
			//cout<<"Vertex found at "<<i<<" ---"<<line<<endl;
			vector<std::string> tokens;
			split(tokens, line, is_any_of(" "));
			for(int j=0; j<tokens.size();j++){
				//If current vertex size is bigger, increment it.
				if(nVertex < atoi(tokens[1].c_str())){
					nVertex++;
					//std::cout << '"' << tokens[j] << '"' << '\n';
				}
			}
			//Save the connected nodes in a vector that stores all connections.
			//The -1 is because nVertex starts in 1 due to the size of array below.
			Vertex[nVertex-1].push_back(tokens[2]);
		}
	}
	//Calculate starting colors.
	nColors = nVertex + 1;
	cout<<"****************************************************"<<endl;
	cout<<"Generating population."<<endl;
	cout<<"****************************************************"<<endl;
	cout<<"Parameters: "<<endl;
	cout<<"Population size: \t" <<pobSize<<endl;
	cout<<"Number of vertex: \t"<<nVertex<<endl;
	cout<<"Number of colors: \t"<<nColors<<endl;
	cout<<"Chromosome length: \t"<<lCrom<<endl;
	cout<<"Crossover rate: \t"<<crossRate<<endl;
	cout<<"Mutation rate: \t\t"<<mutRate<<endl;
	cout<<"Generations: \t\t"<<iterSize<<endl;
	//Generate initial population.
	TIndividual *population[pobSize];
	for(int i=0;i<pobSize;i++){
		population[i] = (TIndividual *)malloc(sizeof(TIndividual));
		genPop(population[i]);
	}
	printPop(population);
	bestIndividual->genes = (int *)malloc(pobSize * sizeof(int));
	cout<<endl<<endl;
	//Evaluation.
	evalPop(population, pobSize, sumAdapt);
	//Evolutive loop.
	for(int i=0; i<iterSize; i++){
		//Selection.
		selectPop(population, pobSize);
		//Crossover.
		reproduction(population, pobSize, lCrom, crossRate);
		//Mutation.
		mutPop(population, pobSize, lCrom, mutRate);
		//Evaluation.
		if(evalPop(population, pobSize, sumAdapt)){
			cout<<"Conflicts not detected, reducing colors"<<endl;
			iterVal = i;
			nColors--;
		}
	}
	printPop(population);
	//Results.
	cout<<endl<<"****************************************************"<<endl;
	cout<<"****************************************************"<<endl;
	cout<<"Best chromosome at position: "<<bestPosition+1<<", value: ";
	int bChain[pobSize];
	for(int j=0;j<lCrom;j++){
		cout<<bestIndividual->genes[j];
		bChain[j] = bestIndividual->genes[j];
	}
	cout<<endl<<"Color number: " << colorNumber<<endl;
	cout<<endl<<"Iteration: "<<iterVal<<endl;
	cout<<endl;
	cout<<"****************************************************"<<endl;
	cout<<"****************************************************"<<endl;
	return 0;
}

vector <std::string> readFile(const char *filename){
	//Reads file to avoid having to type all options.
	std::ifstream infile(filename);
	vector <std::string> lines;
	std::string line;
	if(infile.is_open()){
		while(getline(infile, line)){
			lines.push_back(line);
		}
		cout<<line<<endl;
	}
	return lines;
}

bool IsNum(char c){
    if(isdigit(c) || c=='.')
        return false;
    else
        return true;
}

int calcChromLength(int nColors, float tol){
	int lCrom = 0;
	//The chromosome length is now the number of vertices found.
	lCrom = nVertex;
	//***************************************************
	//TO FIX
	//lCrom = ceil(log((1 + xMax - xMin) / tol) / log(2));
	//***************************************************
	return lCrom;
}

void genPop(struct TIndividual *population){
	//Allocate space for every element inside structure.
	population->genes = (int *)malloc(pobSize * sizeof(int));
	//Iterate through each individual
	for(int j=0;j<lCrom;j++){
			population->genes[j] = rand()%nColors+1;
	}
	//Get each of the genes adaptation value.
	population->adaptation = calcAdapt(population, lCrom);
}

void printPop(TIndividual *population[]){
	//Show each of the generated genes.
	for(int i=0;i<pobSize;i++){
		cout<<endl<<"Individual --- "<<i+1<<endl;
		for(int j=0;j<lCrom;j++){
			cout<<population[i]->genes[j]<<",";
		}
	}
}

void printChromosome(struct TIndividual *population){
	for(int j=0;j<lCrom;j++){
		cout<<population->genes[j]<<",";
	}
}

//This function gets the adaptation value and phenotype of a given chromosome.
float calcAdapt(struct TIndividual *population, int lCrom){
	/*cout<<endl<<"****************************************************"<<endl;
	cout<<"Calculating adaptation value."<<endl;
	cout<<"****************************************************"<<endl;*/
	//Phenotype.
	float phen = 0.0;
	//Function value.
	float func = 0.0;
	//Genes.
	int genes[lCrom];
	for(int j=0;j<lCrom;j++){
		genes[j] = 0;
		genes[j] = population->genes[j];
	}
	//phen = decode(genes, lCrom, nColors);
	//cout<<"Phenotype: "<<phen<<endl;
	//Evaluate phenotype. Given that this is a graph problem,
	//evaluate compared to cost function.
	func = f(population);
	return func;
}

float decode(int genes[], int lCrom, int nColors){
	float decoded = 0.0;
	decoded = bin_ent(genes, lCrom) / (pow(2, lCrom) - 1);
	//**************************************************
	//TO FIX
	//**************************************************
	//decoded = xMin + (xMax - xMin) * decoded;
	return decoded;
}

int bin_ent(int genes[], int lCrom){
	int d=0, pot = 1;
	cout<<"Chromosome: ";
	for(int i=0;i<lCrom;i++){
		d = d + pot * genes[lCrom - i - 1];
		pot = pot * 2;

		cout<<genes[i];
	}
	cout<<"\tInt value: "<<d<<endl;
	return d;
}

int f(struct TIndividual *population){
	float costFunc = 0.0;
	int conflicts = 0;
	int nonconflicts = 0;
	//Iterate through each of the population genes.
	vector <std::string> problem = readFile("test.col");
		//This variable holds the location of the connected solution.
		vector <string> Vertex[problem.size()];
		int k=0;
		for(uint i=0;i<problem.size()-3;i++){
			std::string line = problem[i+3];
			//If e found, it means its a vertex.
			if(line.find("e") != string::npos){
				//cout<<"Vertex found at "<<i<<" ---"<<line<<endl;
				vector<std::string> tokens;
				split(tokens, line, is_any_of(" "));
				for(int j=0; j<tokens.size();j++){
					//If current vertex size is bigger, increment it.
					if(k < atoi(tokens[1].c_str())){
						//std::cout << '"' << tokens[j] << '"' << '\n';
						k++;
					}
				}
				//Save the connected nodes in a vector that stores all connections.
				//The -1 is because nVertex starts in 1 due to the size of array below.
				Vertex[k-1].push_back(tokens[2]);
			}
		}
	/*for(uint i=0;i<nVertex;i++){
		for(uint j=0;j<Vertex[i].size();j++){
			cout<<Vertex[i][j];
		}
		cout<<endl;
	}*/
	//Iterate through each solution.
	for(uint i=0;i<lCrom;i++){
		//Check if colors are connected.
		for(uint j=0;j<Vertex[i].size();j++){
			//cout<<"Gene["<<i<<"] Value: "<<population->genes[i]<<" --- Vertex["<<i<<","<<atoi(Vertex[i][j].c_str())-1<<"] Value: "<<population->genes[atoi(Vertex[i][j].c_str())-1]<<endl;
			if(population->genes[i] == population->genes[atoi(Vertex[i][j].c_str())-1]){
				//cout<<"Conflict!"<<endl;
				conflicts++;
			}
			else{
				nonconflicts++;
			}
		}
		//cout<<endl;
	}
	costFunc = 1 - (1. * conflicts / (conflicts + nonconflicts));
	//cout<<"Cost: "<<costFunc<<endl;
	//cout<<"Number of conflicts: "<<conflicts<<endl;
	population->conflicts = conflicts;
	return conflicts;
}

bool evalPop(struct TIndividual *population[], int pobSize, float sumAdapt){
	/*cout<<endl<<"****************************************************"<<endl;
	cout<<"Evaluating population."<<endl;
	cout<<"****************************************************"<<endl;*/
	int pos=0;
	float punt_acum = 0.0;
	float best_adapt = 100.0;
	sumAdapt = 0.0;
	for(int i=0;i<pobSize;i++){
		//cout<<"Function adaptation: "<<population[i]->adaptation;
		sumAdapt = sumAdapt + population[i]->adaptation;
		//cout<<"\t\tAcumulated: "<<sumAdapt<<endl;
		//Compare the adaptation of actual value to select the best candidate.
		if(population[i]->adaptation < best_adapt){
			pos = i;
			best_adapt = population[i]->adaptation;
		}
	}
	//cout<<endl<<"Position of the best individual: "<<pos+1<<endl;
	for(int i=0;i<pobSize;i++){
		population[i]->rel_val = population[i]->adaptation / sumAdapt;
		population[i]->acc_val = population[i]->rel_val + punt_acum;
		punt_acum = punt_acum + population[i]->rel_val;
	}
	bestPos = pos;
	if(population[bestPos]->conflicts == 0){
		//Store the values of the best individual for safekeeping.
		bestIndividual->genes = population[bestPos]->genes;
		bestIndividual->adaptation = population[bestPos]->adaptation;
		bestIndividual->conflicts = population[bestPos]->conflicts;
		bestPosition = bestPos;
		colorNumber = nColors;
		cout<<"****************************"<<endl;
		for(int j=0;j<lCrom-1;j++){
			cout<<bestIndividual->genes[j]<<",";
		}
		cout<<bestIndividual->adaptation<<endl;
		cout<<bestIndividual->conflicts<<endl;
		cout<<"****************************"<<endl;
		return true;
	}else{
		return false;
	}
}

void selectPop(struct TIndividual *population[], int pobSize){
	/*cout<<endl<<"****************************************************"<<endl;
	cout<<"Initiating selection process."<<endl;
	cout<<"****************************************************"<<endl;*/
	int sel_super_[pobSize];
	float prob;
	int pos_super;
	TIndividual *pob_aux[pobSize];
	for(int i=0;i<pobSize;i++){
		//Generate random numbers between 0 and 1.
		prob = ((double) rand() / (RAND_MAX));
		pos_super = 0;
		while((prob > population[pos_super]->acc_val)&&(pos_super < pobSize)){
			pos_super++;
		}
		sel_super_[i] = pos_super;
	}
	//Generation of intermediate population.
	for(int i=0;i<pobSize;i++){
		pob_aux[i] = population[sel_super_[i]];
	}
	//Initialize population.
	//******************************************************
	//******************************************************
	//for(int i=0;i<pobSize;i++)
	//	genPop(population[i]);	//CHECK IF THIS IS WHAT MUST BE DONE.
	//******************************************************
	//******************************************************
	for(int i=0;i<pobSize;i++){
		population[i] = pob_aux[i];
	}
}

void reproduction(struct TIndividual *population[], int pobSize, int lCrom, float crossRate){
	/*cout<<endl<<"****************************************************"<<endl;
	cout<<"Initiating crossover."<<endl;
	cout<<"****************************************************"<<endl;*/
	int selCross[pobSize];
	int num_sel_cross = 0;
	float prob;
	int punto_cross;
	TIndividual *hijo1 = (TIndividual *)malloc(sizeof(TIndividual));
	//Allocate space for every element inside structure.
	hijo1->genes = (int *)malloc(pobSize * sizeof(int));
	TIndividual *hijo2 = (TIndividual *)malloc(sizeof(TIndividual));
	//Allocate space for every element inside structure.
	hijo2->genes = (int *)malloc(pobSize * sizeof(int));

	//Selecting individuals to cross.
	for(int i=0;i<pobSize;i++){
		//Generate random numbers between 0 and 1.
		prob = ((double) rand() / (RAND_MAX));
		if(prob<crossRate){
			selCross[num_sel_cross] = i;
			num_sel_cross++;
		}
	}
	if((num_sel_cross%2) == 1){
		num_sel_cross--;
	}
	//Cross individuals at random point.
	punto_cross = ((double) rand() / (RAND_MAX));
	for(int i=0;i<num_sel_cross;i=i+2){
		crossPop(population[selCross[i]], population[selCross[i+1]], hijo1, hijo2, lCrom, punto_cross);
		//New individuals replace the fathers.
		population[selCross[i]] = hijo1;
		population[selCross[i+1]] = hijo2;
	}
}

void crossPop(struct TIndividual *population, struct TIndividual *population2, struct TIndividual *hijo1, struct TIndividual *hijo2, int lCrom, int punto_cross){

		for(int i=0;i<punto_cross;i++){
			hijo1->genes[i] = population->genes[i];
			hijo2->genes[i] = population2->genes[i];
		}
		for(int i=punto_cross;i<lCrom;i++){
			hijo1->genes[i] = population2->genes[i];
			hijo2->genes[i] = population->genes[i];
		}
		hijo1->adaptation = calcAdapt(hijo1, lCrom);
		hijo2->adaptation = calcAdapt(hijo2, lCrom);

}

void mutPop(struct TIndividual *population[], int popSize, int lCrom, float mutRate){
	bool mutated = false;
	float prob;
	for(int i=0;i<pobSize;i++){
		mutated = false;
		for(int j=0;j<lCrom;j++){
			prob = ((double) rand() / (RAND_MAX));
			//Mutate genes with prob<mutRate.
			if(prob<mutRate){
				//population[i]->genes[j] = not(population[i]->genes[j]);
				population[i]->genes[j] = rand()%nColors;
				mutated = true;
			}
		}
		if(mutated)
			population[i]->adaptation = calcAdapt(population[i], lCrom);
	}
}
